-- libedit.ex
-- ZIG Library Editor version 0.91
-- Copyright 1999, Jacob Hammond.
-- ZIG webpage: http://surf.to/zig
-- zig16@hotmail.com

include file.e
include get.e
include machine.e
include misc.e
include wildcard.e

global atom
            Scroll_SizeX, Scroll_SizeY, Scroll_TextC, Scroll_BackC,
            Scroll_PosX, Scroll_PosY, Scroll_HLTitleC, Scroll_BorderC    

global sequence LibNames, Config, ObjLibs, LibTabs

include zgfx.e
include zutil.e
include zsidefnc.e
include zlibrs.e
include zsie.e

Config = repeat(FALSE, 20)


global procedure Init()
    sequence dx, vs, cl
    atom f
    
    cl = command_line()
    if length(cl) > 2 then
        for z = 3 to length(cl) do
            if compare(cl[z], "-nodetect") = 0 then Config[1] = TRUE
            elsif compare(cl[z], "-noega") = 0 then Config[2] = TRUE
            elsif find('?', cl[z]) != 0 then
                puts(1, "libedit: Valid commandline options:\n")
                puts(1, "-nodetect - don't try to set the graphics mode\n")
                puts(1, "-noega    - don't use EGA text mode\n")
                abort(0)
            end if
        end for
    end if                
    
    if graphics_mode(3) and not Config[1] then
        puts(1, "libedit: Error entering mode 3 (80x25x16)!\n")
        puts(1, "         This is probably due to a configuration or a detection problem\n")
        puts(1, "         rather than a hardware problem (which would be extremely rare,\n")
        puts(1, "         in this case). Please try again, use the commandline option\n")
        puts(1, "         -nodetect if neccesary.\n")
        abort(1)
    end if
    
    toggle_blinking(0)

    puts(1, "ZIG Library Editor, version 0.91\n")
    puts(1, "Attempting to locate ZIG version descriptor...")
    dx = dir("zig_ver.inf")
    if length(dx) then
        puts(1, "found.\n")
        f = open(dx[1][1], "r")
        vs = get(f)
        vs = vs[2]
        close(f)
        printf(1, "Descriptor says this is ZIG version %d.%d%d.\n", vs)
    else
        puts(1, "\nError: couldn't locate file zig_ver.inf!\n")
    end if
    puts(1, "Copyright 1997-99, Jacob Hammond.\n")
    puts(1, "Press any key to continue...")
    f = wait_key()
                
    cursor(NO_CURSOR)
                
    if length(dir("zig_libs.inf")) then
        f = open("zig_libs.inf", "r")
    else
        f = open("zig_libs.inf", "w")
        if length(dir("stdlib.olf")) then
            puts(f, "stdlib.olf")
        else puts(f, "NULL")
        end if
        close(f)
        f = open("zig_libs.inf", "r")
    end if
    
    LibNames = {}
    LibTabs = {}
    dx = {}
    while TRUE do
        dx = gets(f)
        dx = dx[1..length(dx) - 1]
        puts(1, dx & "\n")
        if not compare(dx, "NULL") then
            exit
        end if
        if length(dir(dx)) then
            LibNames = append(LibNames, dx)
            LibTabs = append(LibTabs, TRUE)
        end if
    end while
    close(f)

    Scroll_SizeX = 16
    Scroll_SizeY = 50
    Scroll_PosX = 5
    Scroll_PosY = 10
    Scroll_TextC = 14
    Scroll_BackC = 1
    Scroll_HLTitleC = 15
    Scroll_BorderC = 9    
    
    ObjLibs = {}
    ObjLibs = LoadLibraries(ObjLibs, LibNames)
    
end procedure

function ColorDialog(atom CColorF, atom CColorB)
    atom Cx, Cy, KRef, OldX, OldY, Input
    sequence Draw, SS
    Draw = repeat(repeat(0, 32), 16)
    
    SS = SaveScreen()
    Box(3, 10, 20, 50, 1, "Choose a Color")
        
    for x = 0 to 15 do
        for y = 0 to 15 do
            Draw[y + 1][((x + 1) * 2) - 1] = 254
            Draw[y + 1][(x + 1) * 2] = ColAttrib(x, y)
        end for
    end for
    
    display_text_image({4, 11}, Draw)
    
    Box(4, 28, 6, 38, 0, "Sample:")
    
    color(14, 1)
    position(8, 28)
    puts(1, "Enter - Accept")
    position(9, 28)
    puts(1, "  ESC - Cancel")
        
    KRef = ON
    
    Cx = CColorB
    Cy = CColorF
    
    while TRUE do
        
        Input = get_key()
        
        OldX = Cx
        OldY = Cy
        
        if Input = KEY_UP then
            Cx = Cx - 1
        elsif Input = KEY_DN then
            Cx = Cx + 1
        elsif Input = KEY_LF then
            Cy = Cy - 1
        elsif Input = KEY_RT then
            Cy = Cy + 1
        elsif Input = KEY_CR then
            ResScreen(SS)
            return ColAttrib(Cy, Cx)
        elsif Input = KEY_ESC then
            ResScreen(SS)
            return ColAttrib(Cy, Cx)
        end if
        
        if Cx != OldX or Cy != OldY or KRef = ON then
            KRef = OFF
            if Cx < 0 or Cx > 15 then
                Cx = OldX
            end if
            if Cy < 0 or Cy > 15 then
                Cy = OldY
            end if
            color(OldY, OldX)
            position(OldX + 4, OldY + 11)
            puts(1, 254)
            color(15, 0)
            position(Cx + 4, Cy + 11)
            puts(1, 197)
            color(Cy, Cx)
            position(5, 29)
            puts(1, {219,178,177,176,32,32})
        end if
    end while
end function

global function GetChar(atom OriChar, atom CC)
    sequence Draw, SS
    atom Counter, CChar, KRef, Cx, Cy, OldX, OldY, OChar, Input, OrigChar
    
    SS = SaveScreen()
    OrigChar = OriChar
    OrigChar = 0    -- bug fixer

    Draw = repeat(repeat(0, 64), 8)
    
    Box(7, 2, 16, 59, 1, "Choose a Character")
    Counter = 0
    
    for x = 0 to 7 do
        for y = 0 to 63 by 2 do
            Draw[x + 1][y + 1] = Counter
            Draw[x + 1][y + 2] = 15
            if Counter = OrigChar then
                Cx = x
                Cy = y
            end if
            Counter = Counter + 1
        end for
    end for
    
    display_text_image({8, 3}, Draw)    
    
    color(14, 1)    
    position(14, 36)
    puts(1, "Enter - Accept")
    position(15, 36)
    puts(1, "  ESC - Cancel")    
    
    OldX = Cx
    OldY = Cy
    KRef = ON
    CChar = OrigChar
    
    while TRUE do
                        
        Input = get_key()

        OldX = Cx
        OldY = Cy
        OChar = CChar
        
        if Input = KEY_UP then
            Cx = Cx - 1
            CChar = CChar - 32
        elsif Input = KEY_DN then
            Cx = Cx + 1
            CChar = CChar + 32
        elsif Input = KEY_LF then
            Cy = Cy - 1
            CChar = CChar - 1
        elsif Input = KEY_RT then
            Cy = Cy + 1
            CChar = CChar + 1
        elsif Input = KEY_CR then
            ResScreen(SS)
            return CChar
        elsif Input = KEY_ESC then
            if OrigChar then
                ResScreen(SS)
                return OrigChar
            end if
        end if
        
        if Cx != OldX or Cy != OldY or KRef = ON then
            
            KRef = OFF
            
            if Cx < 0 or Cx > 7 then
                Cx = OldX
                CChar = OChar
            end if
            if Cy < 0 or Cy > 31 then
                Cy = OldY
                CChar = OChar
            end if

            display_text_image({OldX + 8, OldY + 3}, {{OChar, 15}})
            
            color(14, 0)
            position(Cx + 8, Cy + 3)
            puts(1, 197)
            
            color(14, 1)
            position(8, 36)
            puts(1, "In Current Color: ")
            color(15, 1)
            position(9, 36)
            puts(1, "Char Number: ")
            printf(1, "%3d  ", CChar)
            
            display_text_image({8, 54}, {{CChar, CC}})
        end if
    end while
end function

function MakeLibTable(sequence LibN, sequence v)
    atom cx
    sequence ret
    
    cx = 1
    ret = {}
    
    for z = 1 to length(LibN) do
        if v[z] then
            ret = append(ret, "---" & LibN[z] & "---")
            for a = 1 to length(ObjLibs) do
                if compare(ObjLibs[a][LIB_FROM], LibN[z]) = 0 then
                    ret = append(ret, ObjLibs[a][LIB_NAME])
                end if
            end for
        end if
    end for
    
    return ret

end function

procedure DisplayLibTable(sequence r)
    bk_color(0)
    
    for z = 1 to 25 do
        position(z, 1)
        puts(1, repeat(32, 60))
    end for
    
    for z = 1 to length(r) do
        position(z, 1)
        if r[z][1] = '-' and r[z][3] = '-' and r[z][length(r[z])] = '-' then
            text_color(7)
        else text_color(15)
        end if
        puts(1, r[z])
    end for
end procedure

procedure DrawSelItem(atom s, atom p, sequence l)
    if s then
        text_color(15)
        bk_color(0)
    else
        text_color(14)
        bk_color(1)
    end if

    if p = 1 then position(5, 7) puts(1, "Classification") position(5, 27) puts(1, l[LIB_CLASS])
    elsif p = 2 then position(6, 7) puts(1, "Name") position(6, 27) puts(1, l[LIB_NAME])
    elsif p = 3 then position(7, 7) puts(1, "Default Character") display_text_image({7, 27}, {{l[LIB_CHAR], 15}})
    elsif p = 4 then position(8, 7) puts(1, "Default Color") display_text_image({8, 27}, {{254, l[LIB_COL]}})
    elsif p = 5 then position(9, 7) puts(1, "Object Program") position(9, 27) puts(1, "[program, Enter to edit]")
    end if
    
    for z = 1 to length(l[LIB_PROPS]) do
        if p - ((z - 1) * 3) = 6 then position(12 + ((z - 1) * 5), 7) puts(1, "Property Name") position(12 + ((z - 1) * 5), 27) puts(1, l[LIB_PROPS][z][PROP_NAME])
        elsif p - ((z - 1) * 3) = 7 then position(13 + ((z - 1) * 5), 7) puts(1, "Maximum Value") position(13 + ((z - 1) * 5), 27) printf(1, "%d", l[LIB_PROPS][z][PROP_MAX])
        elsif p - ((z - 1) * 3) = 8 then
            position(14 + ((z - 1) * 5), 7)
            puts(1, "Property Type")
            position(14 + ((z - 1) * 5), 27)
            if l[LIB_PROPS][z][PROP_TYPE] = CHOOSE_NUM then
                puts(1, "Numeric value")
            elsif l[LIB_PROPS][z][PROP_TYPE] = CHOOSE_BOARD then
                puts(1, "Board select")
            elsif l[LIB_PROPS][z][PROP_TYPE] = CHOOSE_CHAR then
                puts(1, "Character select")
            elsif l[LIB_PROPS][z][PROP_TYPE] = CHOOSE_COLOR then
                puts(1, "Color select")
            end if
        end if
    end for

    if p = 99 then
        for z = 1 to length(l[LIB_PROPS]) do
            position(11 + ((z - 1) * 5), 7) text_color(15) printf(1, "Property %d", z) text_color(14)
        end for
    
        text_color(7)
        position(21, 7)
        puts(1, "Library from:")
        position(21, 27)
        puts(1, l[LIB_FROM])
        
        position(22, 7)
        puts(1, "INS/DEL - add/remove user-definable properties")
    end if
end procedure

function GetInputLine(atom x)
    sequence b

    cursor(UNDERLINE_CURSOR)
    position(x, 27)
    puts(1, repeat(32, 20))
    position(x, 27)
    b = gets(0)
    bk_color(1)
    position(x, 27)
    bk_color(1)
    puts(1, repeat(32, 20))
    return b[1..length(b) - 1]
end function

function GetInputNumber(atom x)
    sequence b

    cursor(UNDERLINE_CURSOR)
    position(x, 27)
    puts(1, repeat(32, 20))
    position(x, 27)
    b = get(0)
    bk_color(1)
    position(x, 27)
    bk_color(1)
    puts(1, repeat(32, 20))
    return b[2]
end function

function GetListInput(atom x, atom y, sequence list)
    atom p, op, k, l
    sequence SS

    p = 1
    
    SS = SaveScreen()

    l = length(list[1])
    for z = 2 to length(list) do
        if length(list[z]) > l then l = length(list[z]) end if
    end for
    
    for z = 1 to length(list) do
        if z = p then text_color(0) bk_color(7)
        else text_color(15) bk_color(0) end if
        position((x - 1) + z, y)
        puts(1, " " & list[z] & repeat(' ', l - (length(list[z]) - 1)))
    end for
    
    while TRUE do
    
        op = p
        
        k = get_key()
        if k = KEY_UP then
            p = p - 1
        elsif k = KEY_DN then
            p = p + 1
        elsif k = KEY_ESC or k = KEY_CR then
            exit
        end if
        
        if op != p then
            if p < 1 or p > length(list) then
                p = op
            else
                text_color(15)
                bk_color(0)
                position((x - 1) + op, y)
                puts(1, " " & list[op] & repeat(' ', l - (length(list[op]) - 1)))
                text_color(0)
                bk_color(7)
                position((x - 1) + p, y)
                puts(1, " " & list[p] & repeat(' ', l - (length(list[p]) - 1)))
            end if
        end if
    end while
    
    ResScreen(SS)
    return p - 1
end function

function GetValInput(sequence l, atom p)
    if p = 1 or p = 2 then
        l[p + 1] = GetInputLine(4 + p)
    elsif p = 3 then
        l[LIB_CHAR] = GetChar(0, l[LIB_COL])
    elsif p = 4 then
        l[LIB_COL] = ColorDialog(CGetFore(l[LIB_COL]), CGetBack(l[LIB_COL]))
    elsif p = 5 then
        l[LIB_PROG] = sie_Ed(l[LIB_PROG])
    elsif p = 6 then l[LIB_PROPS][1][PROP_NAME] = GetInputLine(12)
    elsif p = 7 then l[LIB_PROPS][1][PROP_MAX] = GetInputNumber(13)
    elsif p = 8 then
        l[LIB_PROPS][1][PROP_TYPE] =
            GetListInput(14, 27,
            {"Numeric value",
             "Board select",
             "Character select",
             "Color select"})
        position(14, 27)
        bk_color(1)
        puts(1, repeat(' ', 20))
    elsif p = 9 then l[LIB_PROPS][2][PROP_NAME] = GetInputLine(17)
    elsif p = 10 then l[LIB_PROPS][2][PROP_MAX] = GetInputNumber(18)
    elsif p = 11 then
        l[LIB_PROPS][2][PROP_TYPE] =
            GetListInput(19, 27,
            {"Numeric value",
             "Board select",
             "Character select",
             "Color select"})
        position(19, 27)
        bk_color(1)
        puts(1, repeat(' ', 20))
    end if
    
    cursor(NO_CURSOR)
    return l
end function

global function DisplayScroll(sequence Caption, sequence Scroll, atom Sel)
    atom X, PutX, OldX, Key
    sequence TxtScr
    
    TxtScr = save_text_image({1, 1}, {25, 80})
    cursor(NO_CURSOR)
    
    text_color(14)
    bk_color(1)
    
    Box(Scroll_PosX,
        Scroll_PosY,
        (Scroll_PosX - 1) + Scroll_SizeX,
        (Scroll_PosY - 1) + Scroll_SizeY,
        Scroll_BackC,
        Caption)

    if Sel then X = Sel else X = 1 end if
    bk_color(Scroll_BackC)
    
    text_color(Scroll_HLTitleC)
    position(Scroll_PosX + (Scroll_SizeX / 2), Scroll_PosY)
    puts(1, "")
    position(Scroll_PosX + (Scroll_SizeX / 2), Scroll_PosY + Scroll_SizeY - 1)
    puts(1, "")

    Key = -2
    while Key != 27 and Key != 13 do
    
        if Key = -2 then OldX = 0 else OldX = X end if
        
        Key = get_key()
        if Key != -1 then
            if Key = KEY_UP then
                X = X - 1
            elsif Key = KEY_DN then
                X = X + 1
            elsif Key = KEY_PGUP then
                X = X - (Scroll_SizeX - 3)
            elsif Key = KEY_PGDN then
                X = X + (Scroll_SizeX - 3)
            end if
        end if
        
        if X != OldX then
            if X < 1 then
                X = 1
            elsif X > length(Scroll) then
                X = length(Scroll)
            end if
            if X != OldX then
                X = X + 1
                text_color(Scroll_TextC)    
                for z = 1 to Scroll_SizeX - 2 do
                    PutX = (z - (Scroll_SizeX / 2)) + (X - 1)
                    if PutX > 0 and PutX < (length(Scroll) + 1) then                        
                        position(Scroll_PosX + z, Scroll_PosY + 1)
                        if length(Scroll[PutX]) != 0 then
                            if Scroll[PutX][1] = '$' then
                                puts(1, repeat(32, Scroll_SizeY - 2))
                                position(Scroll_PosX + z, (Scroll_PosY + ((Scroll_SizeY / 2) - (length(Scroll[PutX]) / 2))) + 1)
                                text_color(Scroll_HLTitleC)                            
                                puts(1, Scroll[PutX][2..length(Scroll[PutX])])
                                text_color(Scroll_TextC)                            
                            elsif Scroll[PutX][1] = '!' then
                                puts(1, repeat(32, Scroll_SizeY - 2))
                                if z = (Scroll_SizeX / 2) then
                                    text_color(Scroll_TextC)
                                else text_color(Scroll_BorderC)
                                end if
                                position(Scroll_PosX + z, Scroll_PosY + 1)
                                puts(1, "  ")
                                text_color(Scroll_HLTitleC)
                                puts(1, Scroll[PutX][find(';', Scroll[PutX]) + 1..length(Scroll[PutX])])
                            else
                                puts(1, Scroll[PutX])
                                puts(1, repeat(32, (Scroll_SizeY - 2) - length(Scroll[PutX])))
                            end if
                        else
                            puts(1, repeat(32, Scroll_SizeY - 2))
                        end if
                    elsif PutX = 0 or PutX = (length(Scroll) + 1) then
                        position(Scroll_PosX + z, Scroll_PosY + 1)
                        text_color(Scroll_BorderC)
                        puts(1, repeat('*', Scroll_SizeY - 2))
                        text_color(Scroll_TextC)
                    else
                        position(Scroll_PosX + z, Scroll_PosY + 1)
                        puts(1, repeat(32, Scroll_SizeY - 2))
                    end if
                end for
                X = X - 1
                --position(Scroll_PosX, Scroll_PosY + (Scroll_SizeY - 6))
                --printf(1, "%3d", X)
            end if
        end if
    end while
    
    display_text_image({1, 1}, TxtScr)
    
    OldX = (Scroll_SizeX / 2)
    PutX = (OldX - (Scroll_SizeX / 2)) + X
    if length(Scroll[PutX]) != 0 then
        if Scroll[PutX][1] = '!' then
            return(Scroll[PutX][2..(find(';', Scroll[PutX]) - 1)])        
        else return {}
        end if
    else return {}
    end if
end function

global function GetFile(sequence mask)
    sequence RealDir, NDir, Chigger, n
    
    if sequence(dir(mask)) then
        RealDir = dir(mask)
        NDir = {}
        Chigger = {}
        for z = 1 to length(RealDir) do
            NDir = append(NDir, RealDir[z][1])
            Chigger = append(Chigger, "!" & num2str(z) & ";" & NDir[z])
        end for
        n = DisplayScroll("Select a file", Chigger, 1)        
        return NDir[str2num(n)]
    else 
        n = DisplayScroll("Select a file", {"(no files)"}, 1)
        return {1}
    end if
end function

function EditLibItem(sequence li)
    atom p, op, k, m
    sequence SS
    
    SS = SaveScreen()
    
    Box(3, 5, 23, 55, 1, "Editing library item: " & li[LIB_NAME])
    
    for z = 1 to 11 do
        DrawSelItem(0, z, li)
    end for
    
    DrawSelItem(0, 99, li)
        
    text_color(15)
    
    p = 1
    op = 999
    m = 5 + (length(li[LIB_PROPS]) * 3)

    while TRUE do
        if op != 999 then op = p end if
        
        k = get_key()
        if k = KEY_UP then
            p = p - 1
        elsif k = KEY_DN then
            p = p + 1
        elsif k = KEY_CR then
            li = GetValInput(li, p)
            DrawSelItem(1, p, li)
        elsif k = KEY_ESC then
            exit
        elsif k = KEY_INS then
            if length(li[LIB_PROPS]) <= 1 then
                li[LIB_PROPS] = append(li[LIB_PROPS], {"Name",1,1})
            end if
            for z = 1 to 11 do
                DrawSelItem(0, z, li)
            end for
            DrawSelItem(0, 99, li)
            m = 5 + (length(li[LIB_PROPS]) * 3)
        elsif k = KEY_DEL then
            if length(li[LIB_PROPS]) != 0 then
                li[LIB_PROPS] = li[LIB_PROPS][1..length(li[LIB_PROPS]) - 1]
            end if
            Box(3, 5, 23, 55, 1, "Editing library item: " & li[LIB_NAME])
            for z = 1 to 11 do
                DrawSelItem(0, z, li)
            end for
            DrawSelItem(0, 99, li)
            m = 5 + (length(li[LIB_PROPS]) * 3)
        end if
        
        if p != op then
            if op = 999 then op = 1 end if
            if p > m then p = m end if
            if p < 1 then p = 1 end if
            DrawSelItem(0, op, li)
            DrawSelItem(1, p, li)
        end if
    end while
    
    ResScreen(SS)     -- sie may have used the buffer, but it doesn't matter
    return li
end function

procedure Main()
    atom x, oldx, k
    sequence r, fnfind
    object ct, junk
    
    ct = 7
    
    clear_screen()
    SB_Clear()
    
    text_color(0)
    bk_color(3)
    position(2, 62)
    puts(1, "ZIG Library Editor")
  
    SB_Text(5, 15, "Loaded libraries:")    
    for z = 1 to length(LibNames) do
        SB_BtnText(5 + z, ct, " " & 48 + z & "  " & LibNames[z])
        if LibTabs[z] then
            text_color(15)
            bk_color(1)
            position(5 + z, 79)
            puts(1, "*")
        end if
        ct = MBColRotate(ct)
    end for

    SB_Text(15, 15, "Commands:")
    SB_BtnText(16, ct, " Enter  Edit object")  ct = MBColRotate(ct)
    SB_BtnText(17, ct, " Ins  Add object")  ct = MBColRotate(ct)
    SB_BtnText(18, ct, " Del  Delete object")  ct = MBColRotate(ct)
    SB_BtnText(19, ct, " 1.." & 48 + length(LibNames) & "  Vis lib#n")  ct = MBColRotate(ct)
    SB_BtnText(20, ct, " N  New library") ct = MBColRotate(ct)
    SB_BtnText(21, ct, " L  Load a library")  ct = MBColRotate(ct)
    SB_BtnText(22, ct, " U  Unload library")  ct = MBColRotate(ct)
    SB_BtnText(23, ct, " S  Save all")  ct = MBColRotate(ct)
    SB_BtnText(24, ct, " ESC  Quit")  ct = MBColRotate(ct)
    
    r = MakeLibTable(LibNames, LibTabs)
    DisplayLibTable(r)
    
    x = 1
    if r[x][1] = '-' and r[x][3] = '-' and r[x][length(r[x])] = '-' then
        x = x + 1 end if

    if x > length(r) then
        x = -1
    end if
        
    ct = TRUE
    while TRUE do
        oldx = x
        
        k = get_key()
        if k = KEY_UP then
            if x != -1 then
                x = x - 1
            end if
        elsif k = KEY_CR then
            if x != -1 then
                for z = 1 to length(ObjLibs) do
                    if compare(ObjLibs[z][LIB_NAME], r[x]) = 0 then
                        ObjLibs[z] = EditLibItem(ObjLibs[z])
                        exit
                    end if
                end for
                r = MakeLibTable(LibNames, LibTabs)
                DisplayLibTable(r)
                ct = TRUE
            end if
        elsif k = KEY_DN then
            if x != -1 then
                x = x + 1
            end if
        elsif k = KEY_ESC then
            abort(0)
        elsif k >= 49 and k <= 48 + length(LibNames) then
            if length(LibNames) != 0 then
                if LibTabs[k - 48] = TRUE then
                    LibTabs[k - 48] = FALSE
                else LibTabs[k - 48] = TRUE
                end if
            
                ct = 2
                for z = 1 to length(LibTabs) do
                    if LibTabs[z] then
                        ct = 3
                    end if
                end for
                if ct = 2 then
                    LibTabs[1] = TRUE
                end if
            
                r = MakeLibTable(LibNames, LibTabs)
                DisplayLibTable(r)
                if x > length(r) then
                    x = length(r)
                end if
            
                for z = 1 to length(LibNames) do
                    text_color(15)
                    bk_color(1)
                    position(5 + z, 79)
                    if LibTabs[z] then
                        puts(1, "*")
                    else puts(1, ' ') 
                    end if
                end for
                ct = TRUE
                if x = length(r) then
                    oldx = length(r) - 1
                end if
            end if
        elsif k = KEY_INS then
            if length(LibNames) != 0 then
                ct = x
                fnfind = {}
                if x = -1 then ct = length(r) end if
                while TRUE do
                    ct = ct - 1
                    if ct < 1 then
                        fnfind = LibNames[1]
                        exit
                    end if
                    if r[ct][1] = '-' and r[ct][3] = '-' and r[ct][length(r[ct])] = '-' then
                        fnfind = r[ct][4..length(r[ct]) - 3]
                        exit
                    end if
                end while
                ct = {}
                position(1, 27)
                text_color(15)
                puts(1, "New object's name?")
                bk_color(0)
                position(2, 27)
                cursor(UNDERLINE_CURSOR)
                puts(1, repeat(32, 20))
                position(2, 27)
                ct = gets(0)
                cursor(NO_CURSOR)
                ct = ct[1..length(ct) - 1]
                if length(ct) != 0 then
                    ObjLibs = append(ObjLibs,
                            {fnfind,"",ct,1,15,{},{{}}})
                    r = MakeLibTable(LibNames, LibTabs)
                    DisplayLibTable(r)
                    ct = TRUE
                else ct = FALSE
                end if
                position(1, 27)
                puts(1, repeat(32, 20))
            end if
        elsif k = KEY_DEL then
            if x != -1 then
                for z = 1 to length(ObjLibs) do
                    if compare(ObjLibs[z][LIB_NAME], r[x]) = 0 then
                        ObjLibs = ObjLibs[1..z - 1] &
                                    ObjLibs[z + 1..length(ObjLibs)]
                        exit
                    end if
                end for
            
                r = MakeLibTable(LibNames, LibTabs)
                DisplayLibTable(r)
                ct = TRUE
            end if
        elsif k = 'L' or k = 'l' then
            fnfind = GetFile("*.olf")
            if length(fnfind) > 1 then
                LibNames = append(LibNames, fnfind)
                LibTabs = append(LibTabs, TRUE)
                ObjLibs = LoadLibraries(ObjLibs, {fnfind})
            end if
            r = MakeLibTable(LibNames, LibTabs)
            DisplayLibTable(r)
            ct = TRUE
        elsif k = 'U' or k = 'u' then
            if length(LibNames) != 0 then
                ct = x
                if x = -1 then ct = length(r) end if
                fnfind = {}
                while TRUE do
                    ct = ct - 1
                    if ct < 1 then
                        fnfind = LibNames[1]
                        exit
                    end if
                    if r[ct][1] = '-' and r[ct][3] = '-' and r[ct][length(r[ct])] = '-' then
                        fnfind = r[ct][4..length(r[ct]) - 3]
                        exit
                    end if
                end while
                for z = 1 to length(LibNames) do
                    if compare(LibNames[z], fnfind) = 0 then
                        LibNames = LibNames[1..z - 1] & LibNames[z + 1..length(LibNames)]
                        LibTabs = LibTabs[1..z - 1] & LibTabs[z + 1..length(LibTabs)]
                        for a = 1 to length(ObjLibs) do
                            if compare(ObjectLibrary[a][LIB_FROM], fnfind) = 0 then
                                ObjLibs = ObjLibs[1..z - 1] &
                                            ObjLibs[z + 1..length(ObjLibs)]
                            end if
                        end for
                        exit
                    end if
                end for
                
                r = MakeLibTable(LibNames, LibTabs)
                DisplayLibTable(r)
                ct = TRUE
            end if
        elsif k = 'S' or k = 's' then
            for z = 1 to length(LibNames) do
                ct = {}
                for a = 1 to length(ObjLibs) do
                    if compare(ObjLibs[a][LIB_FROM], LibNames[z]) = 0 then
                        ct = append(ct, ObjLibs[a])
                        ct[length(ct)] = ct[length(ct)][2..length(ct[length(ct)])]
                    end if
                end for
                k = open(LibNames[z], "w")
                print(k, ct)
                close(k)
            end for
            ct = FALSE
        elsif k = 'N' or k = 'n' then
            ct = {}
            position(1, 27)
            text_color(15)
            bk_color(0)
            puts(1, "Filename for new library?")
            position(2, 27)
            cursor(UNDERLINE_CURSOR)
            text_color(15)
            ct = gets(0)
            position(3, 27)
            puts(1, "Name for new object?")
            position(4, 27)
            text_color(15)
            junk = gets(0)
            cursor(NO_CURSOR)
            ct = ct[1..length(ct) - 1]
            if length(ct) then
                k = open(ct, "w")
                print(k, {ct,"",junk,1,15,{},{{}}})
                close(k)
                LibNames = append(LibNames, ct)
                LibTabs = append(LibTabs, 1)
                ObjLibs = append(ObjLibs,
                    {ct,"",junk,1,15,{},{{}}})
            end if
            if x = -1 then
                x = 2
            end if

            r = MakeLibTable(LibNames, LibTabs)
            DisplayLibTable(r)
            ct = TRUE
        end if
            
        if x != oldx or ct = TRUE then
            ct = FALSE
            
            if oldx != -1 then
                position(oldx, 1)
                text_color(15)
                bk_color(0)
                puts(1, r[oldx])
            end if

            if x < 1 or x > length(r) then
                x = oldx end if
            
            if x != -1 then
                if r[x][1] = '-' and r[x][3] = '-' and r[x][length(r[x])] = '-' then
                    if (x > oldx or x = 1) and x != length(r) then
                        x = x + 1
                    else
                        x = x - 1
                    end if
                end if
            end if
            
            if x > length(r) then
                x = -1
            else
                position(x, 1)
                text_color(14)
                bk_color(1)
                puts(1, r[x])
            end if
        end if
    end while
            
end procedure

Init()
Main()

