-- zworlds.e
-- ZIG world loading/saving operations

include file.e
include get.e
include machine.e

include zgfx.e

global constant 
    -- Exit constants for LoadWorld() and WriteWorld()
    WFO_SUCCESS = 0, WFO_ERROR = 1, WFO_BADFNAME = 2,
    
    CUR_VERSION = "10A"

-- Read a string of l length from file f
global function readstr(atom f, atom l)
    sequence ret
    
    ret = {}
    
    for z = 1 to l do
        ret = append(ret, getc(f))
    end for
    
    return(ret)
end function

global function WriteWorld(sequence Filename)
    atom f, dul
    sequence SS
    
    SS = SaveScreen()
    
    Box(6, 4, 18, 50, 1, "Saving world: " & Filename)
    position(7, 06)
    
    f = open(Filename, "wb")
    
    puts(1, "File opened.")
    position(8, 06)
    
    puts(1, "Writing header, world / board properties.")
    position(9, 06)

    puts(f, "ZIG")
    puts(f, CUR_VERSION[1..3])
    dul = seek(f, 7)
    print(f, {length(Boards)})  -- store as get()able so we can have +255 brs
    
    dul = seek(f, 13)
    for z = 1 to length(Boards) do
        for a = 1 to 16 do
            for b = 1 to 3 do
                puts(f, BoardProps[z][BP_PAL][a][b])
            end for  
        end for
        for a = 1 to 256 do
            for b = 1 to 16 do
                puts(f, BoardProps[z][BP_CHARSET][a][b])
            end for
        end for
    end for
    
    printf(1, "Writing %d board", length(Boards))
    if length(Boards) > 1 then puts(1, "s") end if
    puts(1, "...")
    position(10, 06)
    
    for z = 1 to length(Boards) do
        for a = 1 to BOARD_ELEMS do
            puts(f, Boards[z][a])
        end for
    end for
    
    dul = seek(f, where(f) + 10)
    printf(1, "Writing %3d objects...", length(Objects))
    position(11, 06)
    puts(1, "Writing settings...")
    print(f, Objects)
    puts(f, ' ')
    print(f, WorldProps)
    puts(f, ' ')
    for z = 1 to length(Boards) do
        print(f, BoardProps[z][1..BP_PAL - 1])
        puts(f, ' ')
    end for

    position(13, 06)
    puts(1, "Complete!")
    close(f)
    ResScreen(SS)
    
    return(WFO_SUCCESS)
end function

global function LoadWorld(sequence Filename)
    atom f, Nb, dul
    sequence VerStamp, getout, SS
    
    if Filename[1] = 1 and length(Filename) = 1 then
        return(WFO_ERROR)
    end if
    
    SS = SaveScreen()

    Box(6, 4, 18, 50, 1, "Loading world: " & Filename)
    position(7, 06)
    
    f = open(Filename, "rb")
    
    puts(1, "File opened.")
    position(8, 06)
    
    puts(1, "Reading header, world / board properties: ")
    position(9, 06)

    VerStamp = get_bytes(f, 6)      -- version stamp
    puts(1, "VerStamp ")
    
    dul = seek(f, 7)
    getout = get(f)  -- stored as get()able so we can have +255 brs
    Nb = getout[2][1]
    puts(1, "NumBoards ")
    
    Boards = repeat(OneBoard, Nb)
    
    dul = seek(f, 13)
    WorldProps = repeat(repeat({}, length(W_PROPS)), Nb)
    BoardProps = repeat(repeat({}, length(B_PROPS)), Nb)
    
    for z = 1 to length(Boards) do
        BoardProps[z][BP_PAL] = DefaultPalette
        BoardProps[z][BP_CHARSET] = CharSet
        
        for a = 1 to 16 do
            for b = 1 to 3 do
                BoardProps[z][BP_PAL][a][b] = getc(f)
            end for  
        end for
        
        for a = 1 to 256 do
            for b = 1 to 16 do
                BoardProps[z][BP_CHARSET][a][b] = getc(f)
            end for
        end for
    end for
    puts(1, "BrdProps.")
    
    position(10, 06)
    printf(1, "Reading %d board", length(Boards))
    if length(Boards) > 1 then puts(1, 's') end if
    puts(1, "...")
    position(10, 06)
    
    for z = 1 to length(Boards) do
        Boards[z] = get_bytes(f, 13500)
    end for
    
    dul = seek(f, where(f) + 10)
    position(11, 06)
    puts(1, "Reading objects...")
    Objects = get(f)        Objects = Objects[2]

    position(12, 06)
    puts(1, "Reading settings...")
    WorldProps = get(f)     WorldProps = WorldProps[2]
    for z = 1 to length(Boards) do
        getout = get(f)     getout = getout[2]
        BoardProps[z] = getout & BoardProps[z][BP_PAL..BP_CHARSET]
    end for
    
    ObjectPosTable = repeat(repeat(repeat(0, 3), 80), 25)
    BuildObjectTable(1)
    
    position(13, 06)
    puts(1, "Complete!")
    close(f)
    ResScreen(SS)
    
    return(WFO_SUCCESS)
end function

